import getpass
import logging
from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel, Field, field_serializer, field_validator

from canvy.const import (
    API_KEY_DESC,
    API_KEY_REGEX,
    DEFAULT_PROVIDER_DESC,
    EDU_URL_DESC,
    OLLAMA_MODEL_DESC,
    OPENAI_KEY_DESC,
    STORAGE_PATH_DESC,
    URL_REGEX,
)

logger = logging.getLogger(__name__)


class ModelProvider(StrEnum):
    OLLAMA = "Ollama"
    OPENAI = "OpenAI"


class CanvyConfig(BaseModel):
    canvas_key: str = Field(description=API_KEY_DESC, pattern=API_KEY_REGEX)
    canvas_url: str = Field(description=EDU_URL_DESC, pattern=URL_REGEX)
    openai_key: str = Field(default="", description=OPENAI_KEY_DESC)
    ollama_model: str = Field(default="", description=OLLAMA_MODEL_DESC)
    storage_path: Path = Field(description=STORAGE_PATH_DESC)
    default_provider: ModelProvider = Field(
        ModelProvider.OPENAI, description=DEFAULT_PROVIDER_DESC
    )

    @field_validator("storage_path")
    @classmethod
    def verify_accessible_path(cls, value: Path) -> Path:
        """
        Test if the user can access a given path to prevent compounding files access errors
        """
        if value.exists() and value.owner() != getpass.getuser():
            e = f"Path {value} exists but we don't have permission to access it"
            raise ValueError(e)
        try:
            value.mkdir(parents=True, exist_ok=True)
        except PermissionError as e:
            logger.error(e)
            raise e
        except Exception as e:
            logger.error(f"Unknown path resolution error: '{e}'")
            raise e
        return value

    @field_serializer("storage_path")
    def serialize_path(self, value: Path) -> str:
        """
        Do this or else it uses __repr__ for some reason
        """
        return str(value)

    @field_serializer("default_provider")
    def serialize_provider(self, value: ModelProvider) -> str:
        """
        Do this or else it turns into array for some reason
        """
        return str(value)


# INFO: Used for the children of modules (ModuleItem)
class ModuleItemType(StrEnum):
    HEADER = "SubHeader"
    PAGE = "Page"
    QUIZ = "Quiz"
    EXTERNAL_TOOL = "ExternalTool"
    EXTERNAL_URL = "ExternalUrl"
    ATTACHMENT = "File"
    DISCUSSION = "Discussion"
    ASSIGNMENT = "Assignment"


# INFO: Used for assignments
class SubmissionType(StrEnum):
    FILE_UPLOAD = "online_upload"
    TEXT_ENTRY = "online_text_entry"
    URL_ENTRY = "online_url"
    RECORDING = "media_recording"
    ANNOTATION = "student_annotation"



import logging.config
import os
import re
from functools import reduce
from pathlib import Path

import toml
from agno.models.base import Model
from agno.models.ollama import Ollama
from agno.models.openai.chat import OpenAIChat
from canvasapi.file import File

from canvy.const import CONFIG_PATH, LOG_FN, LOGGING_CONFIG, OPENAI_MODEL
from canvy.types import CanvyConfig

logger = logging.getLogger(__name__)


def setup_logging() -> None:
    """
    Setup logging using logging config defined in const.py because it's
    quite good
    """
    create_dir(LOG_FN.parent)
    return logging.config.dictConfig(LOGGING_CONFIG)


def better_course_name(name: str) -> str:
    """
    Removes ID numbers next to the given title of the course
    """
    return re.sub(r" ?\((\d,? ?)+\) ?", "", name)


def create_dir(directory: Path) -> None:
    logger.debug(f"Creating directory {directory} if not existing")
    os.makedirs(directory, exist_ok=True)


def get_config(path: Path | None = None) -> CanvyConfig:
    path = path or CONFIG_PATH
    with open(path) as fp:
        logger.debug(f"Retrieving config from {path}")
        config = CanvyConfig(**toml.load(fp))  # pyright: ignore[reportAny]
    return config


def set_config(config: CanvyConfig, dest: Path = CONFIG_PATH) -> None:
    dest = dest if dest else CONFIG_PATH
    with open(dest, "w") as fp:
        logger.debug("Writing config")
        toml.dump(config.model_dump(), fp)


def delete_config(path: Path = CONFIG_PATH):
    os.remove(path)


def download_structured(
    file: File, *dirs: str, storage_dir: Path | None = None, force: bool = False
) -> bool:
    """
    Download a canvasapi File and preserve course structure using directory names

    Args:
        file: File object given by Canvas, can raise various exceptions
        dirs: Series of directory names to make and download file into
        force: Overwrite any previously existing files

    Returns:
        If the file was downloaded
    """
    download_dir = Path(storage_dir or get_config().storage_path).expanduser()
    path: Path = reduce(lambda p, q: p / q, [download_dir, *map(Path, dirs)])
    filename: str = file.filename  # pyright: ignore[reportAny]
    file_path: Path = path / filename
    create_dir(path)
    if not file_path.is_file() or force:
        logger.info(f"Downloading {filename}{'(forced)' * force} into {file_path}")
        try:
            file.download(file_path)  # pyright: ignore[reportUnknownMemberType]
            return True
        except Exception as e:
            logger.warning(
                f"Tried to download {filename} but we likely don't have access ({e})"
            )
            return False
    else:
        logger.info(f"{filename} already present, skipping")
        return False


def provider(config: CanvyConfig) -> Model:
    """
    Get the preferred model provider from the config, default is OpenAI because
    lazy people. Implemented config check to prevent ambiguous errors

    Args:
        config: Config with provider

    Returns:
        Model: Agno model type to use, of which there are many
    """
    e = "Model set to {} but the key isn't in the config"
    if config.default_provider == "OpenAI":
        if key := config.openai_key:
            return OpenAIChat(id=OPENAI_MODEL, api_key=key)
        else:
            raise ValueError(e.format("OpenAI"))
    # elif config.default_provider == "Ollama":
    else:  # noqa: PLR5501
        if model := config.ollama_model:
            return Ollama(id=model)
        else:
            raise ValueError(e.format("Ollama"))



import logging
import os
import sys
from pathlib import Path

import typst
from agno.document.base import Document
from agno.document.reader.pdf_reader import PDFReader

from canvy.const import (
    AGENT_DESCRIPTION,
    AGENT_INSTRUCTIONS,
    OPENAI_EMBEDDINGS,
    PROBLEM_SHEET_1,
    PS_DIRNAME,
    STOP_WORDS,
)
from canvy.types import CanvyConfig
from canvy.utils import create_dir, provider

logger = logging.getLogger(__name__)


def validate_typst(content: str) -> tuple[bool, bytes]:
    try:
        logger.info(f"content: {content}")
        converted = typst.compile(content.encode("utf-8"))
        return (True, converted)
    except Exception as e:
        logger.warning(f"Shit doesn't compile: {e}")
        return (False, str(e).encode("utf-8"))


def make_problem_sheet(config: CanvyConfig):
    def make_problem_sheet(
        file_name: str, class_name: str, title: str, content: str
    ) -> str:
        """
        Produce a problem sheet for the user by recycling content from relevant slides
        and prior knowledge to make engaging yet conformant problems for revision. Focus
        heavily on what the user is asking for.

        Args:
            file_name: File name ending in .pdf
            class_name: Name of the current module or class the topic falls under
            title: Title of the problem sheet
            content: Body of the problem sheet - use normal markdown with many headings

        Return:
            Result
        """
        res, body = validate_typst(
            f"""r
    {PROBLEM_SHEET_1.format(class_name=class_name, student="Mike Hockurts", title=title)}
    {content.replace("#", "=")}
        """
        )
        if not res:
            return body.decode("utf-8")
        sheets_dir = config.storage_path / PS_DIRNAME
        create_dir(sheets_dir)
        with open(sheets_dir / file_name, "wb") as fp:
            fp.write(body)
        logger.info("File written successfully")
        return "Done"

    return make_problem_sheet


def canvas_files(config: CanvyConfig):
    def canvas_files() -> str:
        """
        Produce the local directory of Canvas files / slides so that we can extract text
        from the appropriate files

        Returns:
            Dictionary of {file names -> relative paths}
        """
        return str(
            {
                fn: Path(dir).relative_to(config.storage_path) / fn
                for dir, _, fns in os.walk(config.storage_path)
                for fn in fns
            }
        )

    return canvas_files


def retrieve_knowledge(config: CanvyConfig, queue: list[Document]):
    def retrieve_knowledge(pdf_rel_path: Path):
        """
        Retrieve knowledge which will be processed and added to your knowledge base.

        Args:
            pdf_rel_path: Relative path given by other tools pointing to a PDF file
            queue:

        Returns:
            Confirmation
        """
        queue.extend(PDFReader().read(config.storage_path / pdf_rel_path))
        return "Done. You will now receive knowledge from god."

    return retrieve_knowledge


def teacher(config: CanvyConfig) -> None:
    """
    Basically talk to chatgpt but it can discover about everything in the files downloaded
    through the download tool
    """
    from agno.agent.agent import Agent
    from agno.embedder.openai import OpenAIEmbedder
    from agno.knowledge.pdf import PDFKnowledgeBase
    from agno.tools.duckduckgo import DuckDuckGoTools
    from agno.vectordb.qdrant.qdrant import Qdrant
    from agno.vectordb.search import SearchType

    new_knowledge_queue: list[Document] = []
    agent = Agent(
        model=provider(config),
        description=AGENT_DESCRIPTION,
        instructions=AGENT_INSTRUCTIONS,
        knowledge=PDFKnowledgeBase(
            path=config.storage_path,
            vector_db=Qdrant(
                collection="canvas-files",
                path=str(config.storage_path / ".vector_db"),
                search_type=SearchType.hybrid,
                embedder=OpenAIEmbedder(
                    id=OPENAI_EMBEDDINGS, api_key=config.openai_key
                ),
            ),
        ),
        tools=[
            DuckDuckGoTools(),
            canvas_files(config),
            retrieve_knowledge(config, new_knowledge_queue),
            make_problem_sheet(config),
        ],
        show_tool_calls=True,
        add_history_to_messages=True,
        num_history_runs=3,
        read_chat_history=True,
        markdown=True,
    )
    while True:
        user_input = input(">>> ")
        if user_input in STOP_WORDS:
            sys.exit(0)
        agent.print_response(user_input)  # pyright: ignore[reportUnknownMemberType]
        if new_knowledge_queue and agent.knowledge is not None:
            logger.info("Adding new knowledge")
            agent.knowledge.load_documents(new_knowledge_queue, skip_existing=True)
            new_knowledge_queue.clear()



import logging
from datetime import datetime

from canvasapi.assignment import Assignment
from canvasapi.canvas import Canvas
from canvasapi.grade_change_log import GradeChangeEvent

logger = logging.getLogger(__name__)


# INFO: Postpone but not delete code because idk why I can't access these marks


def grades_by_course(canvas: Canvas) -> dict[str, float | None]:
    return {
        course.course_code: calculate_grading(assignment)
        for course in canvas.get_courses()
        for assignment in course.get_assignments()
    }


def grades(canvas: Canvas) -> dict[str, dict[str, float | None]]:
    return {
        course.course_code: {
            assignment.name: calculate_grading(assignment)
            for assignment in course.get_assignments()
        }
        for course in canvas.get_courses()
    }


def calculate_grading(assignment: Assignment) -> float | None:
    def grade_change_value(event: GradeChangeEvent) -> float:
        t = datetime.fromisoformat(event.created_at)
        return t.timestamp()

    logger.info("Getting grade change events")
    try:
        grade_events = list(assignment.get_grade_change_events())
    except Exception as e:
        logger.warning(f"Error fetching grade events: {e}")
        return None
    if not grade_events:
        logger.warning(f"Grade events for {assignment.name} are empty!")
        return None
    grade_events.sort(key=grade_change_value)
    last_event = grade_events.pop()
    return last_event.grade_after / assignment.points_possible



import logging
import sys
from getpass import getpass
from pathlib import Path

from canvasapi.canvas import Canvas, Course
from canvasapi.requester import ResourceDoesNotExist
from pydantic import ValidationError
from rich import print as pprint
from rich.prompt import Confirm, Prompt
from typer import Typer

from canvy.const import (
    CONFIG_PATH,
    DEFAULT_DOWNLOAD_DIR,
)
from canvy.types import CanvyConfig, ModelProvider
from canvy.utils import (
    better_course_name,
    create_dir,
    delete_config,
    get_config,
    setup_logging,
)

cli = Typer()
logger = logging.getLogger(__name__)


def add_https(url: str):
    return url if url.startswith(("https://", "https://")) else f"https://{url}"


def requires_config() -> CanvyConfig:
    try:
        config = get_config()
        return config
    except FileNotFoundError:
        from canvy.utils import set_config

        choice = Confirm.ask("Config file doesn't exist, create one?")
        if not choice:
            sys.exit(1)
        url = add_https(input("Canvas URL: "))
        api_key_url = f"{url}/profile/settings"
        config = CanvyConfig(
            canvas_url=url,
            canvas_key=getpass(f"Canvas API Key ({api_key_url}): "),
            storage_path=Path(input("Store path (optional): ") or DEFAULT_DOWNLOAD_DIR),
            openai_key=getpass("Open AI Key (optional): "),
            ollama_model=input("Ollama model (optional): "),
            default_provider=ModelProvider.OPENAI,
        )
        set_config(config)
        return requires_config()  # XXX: Might be dangerous :smirking_cat:
    except ValidationError as e:
        pprint(f"Input values are incorrect: {e}")
    except (EOFError, KeyboardInterrupt):
        pprint("\n[bold red]Closing[/bold red]..")
    except Exception as e:
        pprint(f"Either the config is messed up or the internet is: {e}")
        user_choice = Confirm.ask("Destroy config file?")
        if user_choice:
            delete_config()
    sys.exit(1)


def requires_canvas() -> tuple[Canvas, CanvyConfig]:
    config = requires_config()
    canvas = Canvas(config.canvas_url, config.canvas_key)
    return canvas, config


@cli.command(short_help="Download files from Canvas")
def download(*, force: bool = False):
    from canvy.scripts import download

    canvas, config = requires_canvas()
    try:
        count = download(canvas, config.canvas_url, force=force)
        pprint(f"[bold]{count}[/bold] new files! :speaking_head: :fire:")
    except (KeyboardInterrupt, EOFError):
        pprint("[bold red]Download stopping[/bold red]...")
        sys.exit(0)
    except ResourceDoesNotExist as e:
        pprint(f"We likely don't have access to courses no more :sad_cat:: {e}")
        sys.exit(1)


@cli.command(short_help="Use an assistant to go through the files")
def teacher():
    from canvy.scripts import teacher

    config = requires_config()
    try:
        teacher(config)
    except (KeyboardInterrupt, EOFError):
        pprint("[bold red]Program exiting[/bold red]...")
        sys.exit(0)


@cli.command(short_help="List available courses")
def courses(*, detailed: bool = False):
    canvas, _ = requires_canvas()
    try:
        courses: list[Course] = list(canvas.get_courses(enrollment_state="active"))
        if detailed:
            from rich.console import Console
            from rich.table import Table

            table = Table(title="Courses")
            table.add_column("No. Students", style="bold green")
            table.add_column("Title", style="bold")
            table.add_column("Creation date")
            table.add_column("Start date")
            for course in courses:
                table.add_row(
                    getattr(course, "total_students", ""),
                    better_course_name(course.name),
                    course.created_at,
                    getattr(course, "start_at", ""),
                )
            console = Console()
            console.print(table)
        else:
            for course in courses:
                print(f"{course}")  # noqa: T201
    except ResourceDoesNotExist as e:
        pprint(f"We probably don't have access to this course: {e}")
    except Exception as e:
        pprint(f"Unknown error: {e}")


# TODO: Test ts
@cli.command(short_help="Edit config")
def edit_config():
    from canvy.utils import set_config

    current = requires_config()
    try:
        new = CanvyConfig(
            canvas_url=add_https(
                Prompt.ask("Canvas URL: ", default=current.canvas_url)
            ),
            canvas_key=Prompt.ask(
                "Canvas API Key: ",
                show_default=False,
                default=current.canvas_key,
                password=True,
            ),
            storage_path=Path(Prompt.ask("Store path: ", default=current.storage_path)),
            openai_key=Prompt.ask(
                "Open AI Key: ",
                show_default=False,
                default=current.openai_key,
                password=True,
            )
            or "",
            ollama_model=Prompt.ask("Ollama model: ", default=current.ollama_model),
            default_provider=ModelProvider(
                Prompt.ask("Default provider: ", choices=list(ModelProvider))
            ),
        )
        set_config(new)
    except Exception as e:
        pprint(f"[bold red]Bad config[\bold  red]: {e}")


# @cli.command(short_help="Get grades for each course and assignment")
# def grades(*, course_only: bool = False):
#     from canvy.scripts import grades
#
#     canvas, _ = requires_canvas()
#     try:
#         stuff = grades_by_course(canvas) if course_only else grades(canvas)
#         pprint(stuff)
#     except ResourceDoesNotExist as e:
#         pprint(f"We probably don't have access to this course: {e}")
#     except Exception as e:
#         pprint(f"Unknown error: {e}")


@cli.command(short_help="Set up config to use the rest of the tool")
def set_config(  # noqa: PLR0913
    canvas_url: str | None = None,
    canvas_key: str | None = None,
    storage_path: Path | None = None,
    openai_key: str | None = None,
    ollama_model: str | None = None,
    default_provider: ModelProvider | None = None,
):
    from canvy.utils import set_config

    try:
        url = add_https(canvas_url or input("Canvas URL -> https://"))
        api_key_url = f"{url}/profile/settings"
        config = CanvyConfig(
            canvas_url=url,
            canvas_key=canvas_key or getpass(f"Canvas API Key ({api_key_url}): "),
            storage_path=storage_path or DEFAULT_DOWNLOAD_DIR,
            openai_key=openai_key or "",
            ollama_model=ollama_model or "",
            default_provider=default_provider or ModelProvider.OPENAI,
        )
        set_config(config)
    except ValidationError as e:
        pprint(f"Input values are incorrect: {e}")
    except (EOFError, KeyboardInterrupt):
        pprint("\n[bold red]Closing[/bold red]..")


# @cli.callback(invoke_without_command=True)
# def tui():
#     from canvy.tui import run
#     run()


def main():
    create_dir(CONFIG_PATH.parent)
    setup_logging()
    cli()


if __name__ == "__main__":
    main()



from time import monotonic
from typing import ClassVar, override

from textual.app import App, ComposeResult
from textual.binding import BindingType
from textual.containers import HorizontalGroup, VerticalScroll
from textual.reactive import reactive
from textual.timer import Timer
from textual.widgets import Button, Digits, Footer, Header


class TimeDisplay(Digits):
    """
    Sigma
    """

    start_time: reactive[float] = reactive(monotonic)
    time: reactive[float] = reactive(0.0)
    total: reactive[float] = reactive(0.0)
    update_timer: Timer | None = None

    def on_mount(self) -> None:
        self.update_timer = self.set_interval(1 / 60, self.update_time, pause=True)

    def update_time(self) -> None:
        self.time = self.total + (monotonic() - self.start_time)

    def watch_time(self, time: float) -> None:
        minutes, seconds = divmod(time, 60)
        hours, minutes = divmod(minutes, 60)
        self.update(f"{hours:02,.0f}:{minutes:02.0f}:{seconds:05.2f}")

    def start(self) -> None:
        self.start_time = monotonic()
        if self.update_timer is not None:
            self.update_timer.resume()

    def stop(self) -> None:
        if self.update_timer is not None:
            self.update_timer.pause()
        self.total += monotonic() - self.start_time
        self.time = self.total

    def reset(self) -> None:
        self.total = 0
        self.time = 0


class Stopwatch(HorizontalGroup):

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """
        Event handler called when a button is pressed.
        """
        button_id = event.button.id
        time_display = self.query_one(TimeDisplay)
        if button_id == "start":
            time_display.start()
            self.add_class("started")
        elif button_id == "stop":
            time_display.stop()
            self.remove_class("started")
        elif button_id == "reset":
            time_display.reset()

    @override
    def compose(self) -> ComposeResult:
        yield Button("Start", id="start", variant="success")
        yield Button("Stop", id="stop", variant="error")
        yield Button("Reset", id="reset")
        yield TimeDisplay("00:00:00:00")


class StopwatchApp(App[None]):
    CSS = """
    Stopwatch {
        background: $boost;
        height: 5;
        margin: 1;
        min-width: 50;
        padding: 1;
    }

    TimeDisplay {
        text-align: center;
        color: $foreground-muted;
        height: 3;
    }

    Button {
        width: 16;
    }

    #start {
        dock: left;
    }

    #stop {
        dock: left;
        display: none;
    }

    #reset {
        dock: right;
    }

    .started {
        background: $success-muted;
        color: $text;
    }

    .started TimeDisplay {
        color: $foreground;
    }

    .started #start {
        display: none
    }

    .started #stop {
        display: block
    }

    .started #reset {
        visibility: hidden
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        ("a", "add_stopwatch", "Add"),
        ("r", "remove_stopwatch", "Remove"),
    ]

    @override
    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield VerticalScroll(Stopwatch(), Stopwatch(), Stopwatch(), id="timers")

    def action_add_stopwatch(self) -> None:
        new_stopwatch = Stopwatch()
        self.query_one("#timers").mount(new_stopwatch)
        new_stopwatch.scroll_visible()

    def action_remove_stopwatch(self) -> None:
        timers = self.query("Stopwatch")
        if timers:
            timers.last().remove()


def run():
    app = StopwatchApp()
    app.run()



import logging
from pathlib import Path
from typing import Final

from platformdirs import (
    user_config_path,
    user_documents_path,
    user_log_path,
)

from canvy import APP_NAME

logger = logging.getLogger(__name__)

OPENAI_MODEL: Final[str] = "gpt-4.1-mini"
OPENAI_EMBEDDINGS: Final[str] = "text-embedding-3-small"

PROBLEM_SHEET_1: Final[
    str
] = """
#import "@preview/problemst:0.1.2": pset
#show: pset.with(
  class: "{class_name}",
  student: "{student}",
  title: "{title}",
  date: datetime.today(),
)
"""

STOP_WORDS: tuple[str, str, str] = "quit", "/q", "exit"

AGENT_DESCRIPTION: Final[
    str
] = """
    You are an assistant searching through files downloaded from Canvas LMS, user is
    likely a University student.
    """
AGENT_INSTRUCTIONS: Final[list[str]] = [
    "Search your knowledge base for the correct slides and coursework materials",
    "Prefer the information in your knowledge base over the web results.",
    "If the question is better suited for the web, try looking through the canvas files first,\
    and then the web",
]

URL_REGEX: Final[str] = (
    r"[(http(s)?):\/\/(www\.)?a-zA-Z0-9@:%._\+~#=]"
    + r"{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&\/\/=]*)"
)
API_KEY_REGEX: Final[str] = r"\d{4}~[A-Za-z0-9]{64}"
API_KEY_DESC: Final[str] = (
    "API key provided by Canvas which grants access to the API through your account"
)
EDU_URL_DESC: Final[str] = (
    "Institution URL that is used for Canvas, you should have this provided by them."
)
STORAGE_PATH_DESC: Final[str] = "Where to store files"
OPENAI_KEY_DESC: Final[str] = "OpenAI API key to access the GPT models"
ANTHRO_KEY_DESC: Final[str] = "Antrhopic AI API key for the claude models"
OLLAMA_MODEL_DESC: Final[str] = "Ollama model that is going to be used"
DEFAULT_PROVIDER_DESC: Final[str] = "Default model provider, e.g. Ollama, OpenAI etc. "

LOG_FN: Final[Path] = user_log_path(APP_NAME) / "canvy.log"
CONFIG_PATH: Final[Path] = user_config_path(APP_NAME) / "config.toml"
DEFAULT_DOWNLOAD_DIR: Final[Path] = user_documents_path() / APP_NAME
PS_DIRNAME: Final[str] = "Problem Sheets"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {"format": "%(levelname)s: %(message)s"},
        "detailed": {
            "format": "[%(levelname)s|%(module)s|L%(lineno)d] %(asctime)s: %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S%z",
        },
    },
    "handlers": {
        "stderr": {
            "class": "logging.StreamHandler",
            "level": "WARNING",
            "formatter": "simple",
            "stream": "ext://sys.stderr",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "DEBUG",
            "formatter": "detailed",
            "filename": LOG_FN,
            "maxBytes": 10 * 1024**2,
            "backupCount": 3,
        },
    },
    "loggers": {"root": {"level": "DEBUG", "handlers": ["stderr", "file"]}},
}



from canvy.scripts.downloader import download
from canvy.scripts.grades import grades, grades_by_course
from canvy.scripts.teacher import teacher

__all__ = ["download", "grades", "grades_by_course", "teacher"]



# pyright: reportAny=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownMemberType=false
import logging
import re
from collections.abc import Generator
from concurrent.futures import ThreadPoolExecutor

from canvasapi.canvas import Canvas
from canvasapi.course import Course
from canvasapi.exceptions import ResourceDoesNotExist
from canvasapi.file import File
from canvasapi.module import Module, ModuleItem
from canvasapi.page import Page

from canvy.types import ModuleItemType
from canvy.utils import (
    better_course_name,
    download_structured,
)

logger = logging.getLogger(__name__)


def extract_files_from_page(
    canvas: Canvas,
    course: Course,
    module: Module,
    regex: str,
    page: Page,
):
    """
    Use a regex generated from the ID of the course to scrape canvas file links
    and add them to the download queue. We do this because there can be many unmarked
    or arbitrarily organised files on Canvas, depending on the module organiser.
    """
    page_title = getattr(page, "title", "No Title")
    names = [better_course_name(course.name), module.name, page_title]
    if getattr(page, "body", None) is None:
        return
    logging.info(f"Found page: {page}")
    for id in re.findall(regex, page.body):
        if id is None:
            continue
        logger.info(f"Scanned file({id}) from Page({page.page_id})")
        try:
            yield (names, canvas.get_file(id))
        except ResourceDoesNotExist as e:
            logger.warning(f"No access to scrape page: {e}")
        except Exception:
            logger.error(f"Unknown error downloading file {id}")


def module_item_files(
    canvas: Canvas, course: Course, module: Module, regex: str, item: ModuleItem
) -> Generator[tuple[list[str], File], None, None]:
    """
    Process module items into the file queue for downloads
    """
    course_name = better_course_name(course.name)
    if (type := ModuleItemType(item.type)) == ModuleItemType.PAGE:
        page = course.get_page(item.page_url)
        yield from extract_files_from_page(canvas, course, module, regex, page)
    elif type is ModuleItemType.ATTACHMENT:
        file = canvas.get_file(item.content_id)
        names = [course_name, module.name]
        logging.info(f"Found file: {file}")
        yield (names, file)


def download(canvas: Canvas, url: str, *, force: bool = False) -> int:
    """
    Download every file accessible through a Canvas account through courses and modules
    """
    from threading import Lock

    from rich.progress import Progress

    count_lock = Lock()
    download_count = 0

    def safe_download(file: File, paths: list[str]):
        def inner():
            res = download_structured(file, *paths, force=force)
            with count_lock:
                nonlocal download_count
                download_count += res

        return inner

    with Progress() as progress, ThreadPoolExecutor(max_workers=5) as executor:
        user_courses = list(canvas.get_courses(enrollment_state="active"))
        progress_course = progress.add_task("Course", total=len(user_courses))
        progress_module = progress.add_task("Module")
        progress_items = progress.add_task("Downloading files...")
        for course in user_courses:
            progress.update(
                progress_course, description=f"Course: {course.course_code}"
            )
            files_regex = rf"{url}/(?:api/v1/)?courses/{course.id}/files/([0-9]+)"
            for module in (modules := list(course.get_modules())):
                progress.update(
                    progress_module,
                    description=f"Module: {module.name}",
                    total=len(modules),
                )
                for item in (items := list(module.get_module_items())):
                    for paths, file in module_item_files(
                        canvas, course, module, files_regex, item
                    ):
                        progress.update(
                            progress_items,
                            description=f"File: {file.filename}",
                            total=len(items),
                        )
                        executor.submit(safe_download(file, paths))
                    progress.update(progress_items, advance=1)
                progress.update(progress_module, advance=1)
                progress.reset(progress_items)
            progress.update(progress_course, advance=1)
            progress.reset(progress_module)
    return download_count



import logging
from typing import ClassVar, override

from textual.app import App, ComposeResult
from textual.binding import BindingType
from textual.containers import HorizontalGroup, VerticalScroll
from textual.reactive import reactive
from textual.widgets import Button, Digits, Footer, Header

logger = logging.getLogger(__name__)


class Canvy(App[None]):
    BINDINGS: ClassVar[list[BindingType]] = []

    @override
    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        # yield VerticalScroll(Stopwatch(), Stopwatch(), Stopwatch(), id="timers")


def run():
    app = Canvy()
    app.run()
